@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2011 David Robillard <http://drobilla.net>
+  Copyright 2007-2014 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -28,17 +28,25 @@ extern "C" {
 
 #ifdef _WIN32
 #    include <windows.h>
+#    include <direct.h>
 #    define dlopen(path, flags) LoadLibrary(path)
-#    define dlclose(lib) FreeLibrary((HMODULE)lib)
+#    define dlclose(lib)        FreeLibrary((HMODULE)lib)
+#    define unlink(path)        _unlink(path)
+#    define rmdir(path)         _rmdir(path)
 #    ifdef _MSC_VER
 #        define __func__ __FUNCTION__
-#        define INFINITY DBL_MAX + DBL_MAX
-#        define NAN INFINITY - INFINITY
 #        define snprintf _snprintf
 #    endif
+#ifndef INFINITY
+#    define INFINITY DBL_MAX + DBL_MAX
+#endif
+#ifndef NAN
+#    define NAN INFINITY - INFINITY
+#endif
 static inline char* dlerror(void) { return "Unknown error"; }
 #else
 #    include <dlfcn.h>
+#    include <unistd.h>
 #endif
 
 #include "serd/serd.h"
@@ -99,11 +107,10 @@ typedef struct {
 typedef struct {
 	LilvWorld*                world;
 	LilvNode*                 uri;
+	char*                     bundle_path;
 	void*                     lib;
 	LV2_Descriptor_Function   lv2_descriptor;
-#ifdef LILV_NEW_LV2
 	const LV2_Lib_Descriptor* desc;
-#endif
 	uint32_t                  refs;
 } LilvLib;
 
@@ -174,7 +181,7 @@ struct LilvWorldImpl {
 		SordNode* lv2_requiredFeature;
 		SordNode* lv2_symbol;
 		SordNode* lv2_prototype;
-		SordNode* null_uri;
+		SordNode* owl_Ontology;
 		SordNode* pset_value;
 		SordNode* rdf_a;
 		SordNode* rdf_value;
@@ -187,6 +194,7 @@ struct LilvWorldImpl {
 		SordNode* xsd_decimal;
 		SordNode* xsd_double;
 		SordNode* xsd_integer;
+		SordNode* null_uri;
 	} uris;
 	LilvOptions opt;
 };
@@ -274,30 +282,34 @@ LilvScalePoints*   lilv_scale_points_new(void);
 LilvPluginClasses* lilv_plugin_classes_new(void);
 LilvUIs*           lilv_uis_new(void);
 
+LilvNode* lilv_world_get_manifest_uri(LilvWorld* world, LilvNode* bundle_uri);
+
 const uint8_t* lilv_world_blank_node_prefix(LilvWorld* world);
-void           lilv_world_load_file(LilvWorld* world, const char* file_uri);
+
+SerdStatus lilv_world_load_file(LilvWorld*      world,
+                                SerdReader*     reader,
+                                const LilvNode* uri);
+
+SerdStatus
+lilv_world_load_graph(LilvWorld*      world,
+                      SordNode*       graph,
+                      const LilvNode* uri);
 
 LilvUI* lilv_ui_new(LilvWorld* world,
-                    LilvNode* uri,
-                    LilvNode* type_uri,
-                    LilvNode* binary_uri);
+                    LilvNode*  uri,
+                    LilvNode*  type_uri,
+                    LilvNode*  binary_uri);
 
 void lilv_ui_free(LilvUI* ui);
 
-LilvNode*       lilv_node_new(LilvWorld*   world,
-                              LilvNodeType type,
-                              const char*  val);
-LilvNode*       lilv_node_new_from_node(LilvWorld*      world,
-                                        const SordNode* node);
-const SordNode* lilv_node_as_node(const LilvNode* value);
+LilvNode* lilv_node_new(LilvWorld* world, LilvNodeType type, const char* val);
+LilvNode* lilv_node_new_from_node(LilvWorld* world, const SordNode* node);
 
 int lilv_header_compare_by_uri(const void* a, const void* b, void* user_data);
+int lilv_lib_compare(const void* a, const void* b, void* user_data);
 
-int
-lilv_ptr_cmp(const void* a, const void* b, void* user_data);
-
-int
-lilv_resource_node_cmp(const void* a, const void* b, void* user_data);
+int lilv_ptr_cmp(const void* a, const void* b, void* user_data);
+int lilv_resource_node_cmp(const void* a, const void* b, void* user_data);
 
 struct LilvHeader*
 lilv_collection_get_by_uri(const ZixTree* seq, const LilvNode* uri);
@@ -323,6 +335,14 @@ lilv_world_find_nodes_internal(LilvWorld*      world,
                                const SordNode* predicate,
                                const SordNode* object);
 
+SordModel*
+lilv_world_filter_model(LilvWorld*      world,
+                        SordModel*      model,
+                        const SordNode* subject,
+                        const SordNode* predicate,
+                        const SordNode* object,
+                        const SordNode* graph);
+
 #define FOREACH_MATCH(iter) \
 	for (; !sord_iter_end(iter); sord_iter_next(iter))
 
