@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2011 David Robillard <http://drobilla.net>
+  Copyright 2007-2015 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -14,7 +14,7 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#include <assert.h>
+#include <math.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -44,7 +44,7 @@ lilv_node_set_numerics_from_string(LilvNode* val, size_t len)
 	}
 }
 
-/** Note that if @a type is numeric or boolean, the returned value is corrupt
+/** Note that if `type` is numeric or boolean, the returned value is corrupt
  * until lilv_node_set_numerics_from_string is called.  It is not
  * automatically called from here to avoid overhead and imprecision when the
  * exact string value is known.
@@ -93,7 +93,7 @@ lilv_node_new(LilvWorld* world, LilvNodeType type, const char* str)
 	return val;
 }
 
-/** Create a new LilvNode from @a node, or return NULL if impossible */
+/** Create a new LilvNode from `node`, or return NULL if impossible */
 LilvNode*
 lilv_node_new_from_node(LilvWorld* world, const SordNode* node)
 {
@@ -140,29 +140,37 @@ lilv_node_new_from_node(LilvWorld* world, const SordNode* node)
 			world, type, (const char*)sord_node_get_string_counted(node, &len));
 		lilv_node_set_numerics_from_string(result, len);
 		break;
-	default:
-		assert(false);
 	}
 
 	return result;
 }
 
-LILV_API
-LilvNode*
+LILV_API LilvNode*
 lilv_new_uri(LilvWorld* world, const char* uri)
 {
 	return lilv_node_new(world, LILV_VALUE_URI, uri);
 }
 
-LILV_API
-LilvNode*
+LILV_API LilvNode*
+lilv_new_file_uri(LilvWorld* world, const char* host, const char* path)
+{
+	char*    abs_path = lilv_path_absolute(path);
+	SerdNode s        = serd_node_new_file_uri(
+		(const uint8_t*)abs_path, (const uint8_t*)host, NULL, true);
+
+	LilvNode* ret = lilv_node_new(world, LILV_VALUE_URI, (const char*)s.buf);
+	serd_node_free(&s);
+	free(abs_path);
+	return ret;
+}
+
+LILV_API LilvNode*
 lilv_new_string(LilvWorld* world, const char* str)
 {
 	return lilv_node_new(world, LILV_VALUE_STRING, str);
 }
 
-LILV_API
-LilvNode*
+LILV_API LilvNode*
 lilv_new_int(LilvWorld* world, int val)
 {
 	char str[32];
@@ -172,8 +180,7 @@ lilv_new_int(LilvWorld* world, int val)
 	return ret;
 }
 
-LILV_API
-LilvNode*
+LILV_API LilvNode*
 lilv_new_float(LilvWorld* world, float val)
 {
 	char str[32];
@@ -183,8 +190,7 @@ lilv_new_float(LilvWorld* world, float val)
 	return ret;
 }
 
-LILV_API
-LilvNode*
+LILV_API LilvNode*
 lilv_new_bool(LilvWorld* world, bool val)
 {
 	LilvNode* ret = lilv_node_new(world, LILV_VALUE_BOOL,
@@ -193,8 +199,7 @@ lilv_new_bool(LilvWorld* world, bool val)
 	return ret;
 }
 
-LILV_API
-LilvNode*
+LILV_API LilvNode*
 lilv_node_duplicate(const LilvNode* val)
 {
 	if (!val) {
@@ -209,8 +214,7 @@ lilv_node_duplicate(const LilvNode* val)
 	return result;
 }
 
-LILV_API
-void
+LILV_API void
 lilv_node_free(LilvNode* val)
 {
 	if (val) {
@@ -219,8 +223,7 @@ lilv_node_free(LilvNode* val)
 	}
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_equals(const LilvNode* value, const LilvNode* other)
 {
 	if (value == NULL && other == NULL)
@@ -247,8 +250,7 @@ lilv_node_equals(const LilvNode* value, const LilvNode* other)
 	return false; /* shouldn't get here */
 }
 
-LILV_API
-char*
+LILV_API char*
 lilv_node_get_turtle_token(const LilvNode* value)
 {
 	const char* str    = (const char*)sord_node_get_string(value->node);
@@ -285,45 +287,35 @@ lilv_node_get_turtle_token(const LilvNode* value)
 	return result;
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_uri(const LilvNode* value)
 {
 	return (value && value->type == LILV_VALUE_URI);
 }
 
-LILV_API
-const char*
+LILV_API const char*
 lilv_node_as_uri(const LilvNode* value)
 {
-	assert(lilv_node_is_uri(value));
-	return (const char*)sord_node_get_string(value->node);
+	return (lilv_node_is_uri(value)
+	        ? (const char*)sord_node_get_string(value->node)
+	        : NULL);
 }
 
-const SordNode*
-lilv_node_as_node(const LilvNode* value)
-{
-	assert(lilv_node_is_uri(value));
-	return value->node;
-}
-
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_blank(const LilvNode* value)
 {
 	return (value && value->type == LILV_VALUE_BLANK);
 }
 
-LILV_API
-const char*
+LILV_API const char*
 lilv_node_as_blank(const LilvNode* value)
 {
-	assert(lilv_node_is_blank(value));
-	return (const char*)sord_node_get_string(value->node);
+	return (lilv_node_is_blank(value)
+	        ? (const char*)sord_node_get_string(value->node)
+	        : NULL);
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_literal(const LilvNode* value)
 {
 	if (!value)
@@ -333,73 +325,71 @@ lilv_node_is_literal(const LilvNode* value)
 	case LILV_VALUE_STRING:
 	case LILV_VALUE_INT:
 	case LILV_VALUE_FLOAT:
+	case LILV_VALUE_BLOB:
 		return true;
 	default:
 		return false;
 	}
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_string(const LilvNode* value)
 {
 	return (value && value->type == LILV_VALUE_STRING);
 }
 
-LILV_API
-const char*
+LILV_API const char*
 lilv_node_as_string(const LilvNode* value)
 {
 	return value ? (const char*)sord_node_get_string(value->node) : NULL;
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_int(const LilvNode* value)
 {
 	return (value && value->type == LILV_VALUE_INT);
 }
 
-LILV_API
-int
+LILV_API int
 lilv_node_as_int(const LilvNode* value)
 {
-	assert(value);
-	assert(lilv_node_is_int(value));
-	return value->val.int_val;
+	return lilv_node_is_int(value) ? value->val.int_val : 0;
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_float(const LilvNode* value)
 {
 	return (value && value->type == LILV_VALUE_FLOAT);
 }
 
-LILV_API
-float
+LILV_API float
 lilv_node_as_float(const LilvNode* value)
 {
-	assert(lilv_node_is_float(value) || lilv_node_is_int(value));
 	if (lilv_node_is_float(value)) {
 		return value->val.float_val;
-	} else {  // lilv_node_is_int(value)
+	} else if (lilv_node_is_int(value)) {
 		return (float)value->val.int_val;
 	}
+	return NAN;
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_is_bool(const LilvNode* value)
 {
 	return (value && value->type == LILV_VALUE_BOOL);
 }
 
-LILV_API
-bool
+LILV_API bool
 lilv_node_as_bool(const LilvNode* value)
 {
-	assert(value);
-	assert(lilv_node_is_bool(value));
-	return value->val.bool_val;
+	return lilv_node_is_bool(value) ? value->val.bool_val : false;
+}
+
+LILV_API char*
+lilv_node_get_path(const LilvNode* value, char** hostname)
+{
+	if (lilv_node_is_uri(value)) {
+		return lilv_file_uri_parse(lilv_node_as_uri(value), hostname);
+	}
+	return NULL;
 }
