@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2011 David Robillard <http://drobilla.net>
+  Copyright 2007-2015 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -14,17 +14,17 @@
   OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
 
-#define _POSIX_C_SOURCE 1  /* for fileno */
-#define _BSD_SOURCE     1  /* for realpath, symlink */
+#define _POSIX_C_SOURCE 200809L  /* for fileno */
+#define _BSD_SOURCE     1        /* for realpath, symlink */
 
 #ifdef __APPLE__
 #    define _DARWIN_C_SOURCE 1  /* for flock */
 #endif
 
-#include <assert.h>
 #include <ctype.h>
 #include <errno.h>
 #include <stdarg.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -35,8 +35,20 @@
 #    include <io.h>
 #    define F_OK 0
 #    define mkdir(path, flags) _mkdir(path)
+#    if (defined(_MSC_VER) && (_MSC_VER < 1500))
+/** Implement 'CreateSymbolicLink()' for MSVC 8 or earlier */
+BOOLEAN WINAPI
+CreateSymbolicLink(LPCTSTR linkpath, LPCTSTR targetpath, DWORD flags)
+{
+	typedef BOOLEAN (WINAPI* PFUNC)(LPCTSTR, LPCTSTR, DWORD);
+
+	PFUNC pfn = (PFUNC)GetProcAddress(GetModuleHandle(TEXT("kernel32.dll")), "CreateSymbolicLinkA");
+	return pfn ? pfn(linkpath, targetpath, flags) : 0;
+}
+#    endif /* _MSC_VER < 1500 */
 #else
 #    include <dirent.h>
+#    include <limits.h>
 #    include <unistd.h>
 #endif
 
@@ -53,6 +65,12 @@
 #   define PAGE_SIZE 4096
 #endif
 
+void
+lilv_free(void* ptr)
+{
+	free(ptr);
+}
+
 char*
 lilv_strjoin(const char* first, ...)
 {
@@ -68,13 +86,14 @@ lilv_strjoin(const char* first, ...)
 		if (s == NULL)
 			break;
 
-		const size_t this_len = strlen(s);
-		if (!(result = (char*)realloc(result, len + this_len + 1))) {
+		const size_t this_len   = strlen(s);
+		char*        new_result = (char*)realloc(result, len + this_len + 1);
+		if (!new_result) {
 			free(result);
-			LILV_ERROR("realloc() failed\n");
 			return NULL;
 		}
 
+		result = new_result;
 		memcpy(result + len, s, this_len);
 		len += this_len;
 	}
@@ -92,10 +111,10 @@ lilv_strdup(const char* str)
 		return NULL;
 	}
 
-	const size_t len = strlen(str);
-	char*        dup = (char*)malloc(len + 1);
-	memcpy(dup, str, len + 1);
-	return dup;
+	const size_t len  = strlen(str);
+	char*        copy = (char*)malloc(len + 1);
+	memcpy(copy, str, len + 1);
+	return copy;
 }
 
 const char*
@@ -104,6 +123,12 @@ lilv_uri_to_path(const char* uri)
 	return (const char*)serd_uri_to_path((const uint8_t*)uri);
 }
 
+char*
+lilv_file_uri_parse(const char* uri, char** hostname)
+{
+	return (char*)serd_file_uri_parse((const uint8_t*)uri, (uint8_t**)hostname);
+}
+
 /** Return the current LANG converted to Turtle (i.e. RFC3066) style.
  * For example, if LANG is set to "en_CA.utf-8", this returns "en-ca".
  */
@@ -165,7 +190,7 @@ append_var(char* dst, size_t* dst_len, const char* var)
 	}
 }
 
-/** Expand variables (e.g. POSIX ~ or $FOO, Windows %FOO%) in @a path. */
+/** Expand variables (e.g. POSIX ~ or $FOO, Windows %FOO%) in `path`. */
 char*
 lilv_expand(const char* path)
 {
@@ -267,36 +292,33 @@ lilv_copy_file(const char* src, const char* dst)
 {
 	FILE* in = fopen(src, "r");
 	if (!in) {
-		LILV_ERRORF("error opening %s (%s)\n", src, strerror(errno));
-		return 1;
+		return errno;
 	}
 
 	FILE* out = fopen(dst, "w");
 	if (!out) {
-		LILV_ERRORF("error opening %s (%s)\n", dst, strerror(errno));
-		fclose(in);
-		return 2;
+		return errno;
 	}
 
 	char*  page   = (char*)malloc(PAGE_SIZE);
 	size_t n_read = 0;
+	int    st     = 0;
 	while ((n_read = fread(page, 1, PAGE_SIZE, in)) > 0) {
 		if (fwrite(page, 1, n_read, out) != n_read) {
-			LILV_ERRORF("write to %s failed (%s)\n", dst, strerror(errno));
+			st = errno;
 			break;
 		}
 	}
 
-	const int ret = ferror(in) || ferror(out);
-	if (ferror(in)) {
-		LILV_ERRORF("read from %s failed (%s)\n", src, strerror(errno));
+	if (!st && (ferror(in) || ferror(out))) {
+		st = EBADF;
 	}
 
 	free(page);
 	fclose(in);
 	fclose(out);
 
-	return ret;
+	return st;
 }
 
 bool
@@ -355,6 +377,7 @@ lilv_size_mtime(const char* path, off_t* size, time_t* time)
 	struct stat buf;
 	if (stat(path, &buf)) {
 		LILV_ERRORF("stat(%s) (%s)\n", path, strerror(errno));
+		return;
 	}
 
 	if (size) {
@@ -392,7 +415,7 @@ update_latest(const char* path, const char* name, void* data)
 	}
 }
 
-/** Return the latest copy of the file at @c path that is newer. */
+/** Return the latest copy of the file at `path` that is newer. */
 char*
 lilv_get_latest_copy(const char* path, const char* copy_path)
 {
@@ -410,13 +433,23 @@ lilv_get_latest_copy(const char* path, const char* copy_path)
 char*
 lilv_realpath(const char* path)
 {
-#ifdef _WIN32
+#if defined(_WIN32)
 	char* out = (char*)malloc(MAX_PATH);
 	GetFullPathName(path, MAX_PATH, out, NULL);
 	return out;
-#else
+#elif _POSIX_VERSION >= 200809L
 	char* real_path = realpath(path, NULL);
 	return real_path ? real_path : lilv_strdup(path);
+#else
+	// OSX <= 10.5, if anyone cares.  I sure don't.
+	char* out       = (char*)malloc(PATH_MAX);
+	char* real_path = realpath(path, out);
+	if (!real_path) {
+		free(out);
+		return lilv_strdup(path);
+	} else {
+		return real_path;
+	}
 #endif
 }
 
@@ -427,6 +460,9 @@ lilv_symlink(const char* oldpath, const char* newpath)
 	if (strcmp(oldpath, newpath)) {
 #ifdef _WIN32
 		ret = !CreateSymbolicLink(newpath, oldpath, 0);
+		if (ret) {
+			ret = !CreateHardLink(newpath, oldpath, 0);
+		}
 #else
 		ret = symlink(oldpath, newpath);
 #endif
@@ -517,11 +553,18 @@ lilv_dir_for_each(const char* path,
 #else
 	DIR* dir = opendir(path);
 	if (dir) {
-		struct dirent  entry;
-		struct dirent* result;
-		while (!readdir_r(dir, &entry, &result) && result) {
-			f(path, entry.d_name, data);
+		long name_max = pathconf(path, _PC_NAME_MAX);
+		if (name_max == -1) {
+			name_max = 255;   // Limit not defined, or error
 		}
+
+		const size_t   len    = offsetof(struct dirent, d_name) + name_max + 1;
+		struct dirent* entry = (struct dirent*)malloc(len);
+		struct dirent* result;
+		while (!readdir_r(dir, entry, &result) && result) {
+			f(path, entry->d_name, data);
+		}
+		free(entry);
 		closedir(dir);
 	}
 #endif
@@ -536,10 +579,8 @@ lilv_mkdir_p(const char* dir_path)
 		if (path[i] == LILV_DIR_SEP[0] || path[i] == '\0') {
 			path[i] = '\0';
 			if (mkdir(path, 0755) && errno != EEXIST) {
-				LILV_ERRORF("Failed to create %s (%s)\n",
-				            path, strerror(errno));
 				free(path);
-				return 1;
+				return errno;
 			}
 			path[i] = LILV_DIR_SEP[0];
 		}
