@@ -1,5 +1,5 @@
 /*
-  Copyright 2007-2011 David Robillard <http://drobilla.net>
+  Copyright 2007-2015 David Robillard <http://drobilla.net>
   Copyright 2008 Krzysztof Foltman
 
   Permission to use, copy, modify, and/or distribute this software for any
@@ -41,6 +41,7 @@
 #include "lilv/lilv.h"
 #include "../src/lilv_internal.h"
 
+#include "lv2/lv2plug.in/ns/ext/presets/presets.h"
 #include "lv2/lv2plug.in/ns/ext/state/state.h"
 #include "lv2/lv2plug.in/ns/ext/urid/urid.h"
 
@@ -158,6 +159,7 @@ cleanup(void)
 #define TEST_ASSERT(check) do {\
 	test_count++;\
 	if (!(check)) {\
+		assert(false);\
 		error_count++;\
 		fprintf(stderr, "lilv_test.c:%d: error: %s\n", __LINE__, #check);\
 	}\
@@ -179,9 +181,10 @@ struct TestCase {
 #define PREFIX_RDFS "@prefix rdfs: <http://www.w3.org/2000/01/rdf-schema#> .\n"
 #define PREFIX_FOAF "@prefix foaf: <http://xmlns.com/foaf/0.1/> .\n"
 #define PREFIX_DOAP "@prefix doap: <http://usefulinc.com/ns/doap#> .\n"
+#define PREFIX_PSET "@prefix pset: <http://lv2plug.in/ns/ext/presets#> .\n"
 
 #define MANIFEST_PREFIXES PREFIX_LINE PREFIX_LV2 PREFIX_RDFS
-#define BUNDLE_PREFIXES PREFIX_ATOM PREFIX_LINE PREFIX_LV2 PREFIX_RDF PREFIX_RDFS PREFIX_FOAF PREFIX_DOAP
+#define BUNDLE_PREFIXES PREFIX_ATOM PREFIX_LINE PREFIX_LV2 PREFIX_RDF PREFIX_RDFS PREFIX_FOAF PREFIX_DOAP PREFIX_PSET
 #define PLUGIN_NAME(name) "doap:name \"" name "\""
 #define LICENSE_GPL "doap:license <http://usefulinc.com/doap/licenses/gpl>"
 
@@ -211,17 +214,6 @@ cleanup_uris(void)
 
 /*****************************************************************************/
 
-static int
-test_utils(void)
-{
-	TEST_ASSERT(!strcmp(lilv_uri_to_path("file:///tmp/blah"), "/tmp/blah"));
-	TEST_ASSERT(!lilv_uri_to_path("file:/example.org/blah"));
-	TEST_ASSERT(!lilv_uri_to_path("http://example.org/blah"));
-	return 1;
-}
-
-/*****************************************************************************/
-
 static int
 test_value(void)
 {
@@ -249,28 +241,45 @@ test_value(void)
 	TEST_ASSERT(lilv_node_is_int(ival));
 	TEST_ASSERT(lilv_node_is_float(fval));
 
+	TEST_ASSERT(!lilv_node_is_literal(NULL));
 	TEST_ASSERT(!lilv_node_is_literal(uval));
 	TEST_ASSERT(lilv_node_is_literal(sval));
 	TEST_ASSERT(lilv_node_is_literal(ival));
 	TEST_ASSERT(lilv_node_is_literal(fval));
+	TEST_ASSERT(!lilv_node_get_path(fval, NULL));
 
 	TEST_ASSERT(!strcmp(lilv_node_as_uri(uval), "http://example.org"));
 	TEST_ASSERT(!strcmp(lilv_node_as_string(sval), "Foo"));
 	TEST_ASSERT(lilv_node_as_int(ival) == 42);
 	TEST_ASSERT(fabs(lilv_node_as_float(fval) - 1.6180) < FLT_EPSILON);
 
+	LilvNode* loc_abs  = lilv_new_file_uri(world, NULL, "/foo/bar");
+	LilvNode* loc_rel  = lilv_new_file_uri(world, NULL, "foo");
+	LilvNode* host_abs = lilv_new_file_uri(world, "host", "/foo/bar");
+	LilvNode* host_rel = lilv_new_file_uri(world, "host", "foo");
+
+	TEST_ASSERT(!strcmp(lilv_node_as_uri(loc_abs), "file:///foo/bar"));
+	TEST_ASSERT(!strncmp(lilv_node_as_uri(loc_rel), "file:///", 8));
+	TEST_ASSERT(!strcmp(lilv_node_as_uri(host_abs), "file://host/foo/bar"));
+	TEST_ASSERT(!strncmp(lilv_node_as_uri(host_rel), "file://host/", 12));
+
+	lilv_node_free(host_rel);
+	lilv_node_free(host_abs);
+	lilv_node_free(loc_rel);
+	lilv_node_free(loc_abs);
+
 	char* tok = lilv_node_get_turtle_token(uval);
 	TEST_ASSERT(!strcmp(tok, "<http://example.org>"));
-	free(tok);
+	lilv_free(tok);
 	tok = lilv_node_get_turtle_token(sval);
 	TEST_ASSERT(!strcmp(tok, "Foo"));
-	free(tok);
+	lilv_free(tok);
 	tok = lilv_node_get_turtle_token(ival);
 	TEST_ASSERT(!strcmp(tok, "42"));
-	free(tok);
+	lilv_free(tok);
 	tok = lilv_node_get_turtle_token(fval);
 	TEST_ASSERT(!strncmp(tok, "1.6180", 6));
-	free(tok);
+	lilv_free(tok);
 
 	LilvNode* uval_e = lilv_new_uri(world, "http://example.org");
 	LilvNode* sval_e = lilv_new_string(world, "Foo");
@@ -331,6 +340,16 @@ test_value(void)
 
 /*****************************************************************************/
 
+static int
+test_util(void)
+{
+	TEST_ASSERT(!lilv_realpath(NULL));
+
+	return 1;
+}
+
+/*****************************************************************************/
+
 static int discovery_plugin_found = 0;
 
 static void
@@ -570,12 +589,18 @@ test_plugin(void)
 	                           klass_uri));
 	lilv_node_free(rdf_type);
 
+	TEST_ASSERT(!lilv_plugin_is_replaced(plug));
+	TEST_ASSERT(!lilv_plugin_get_related(plug, NULL));
+
 	const LilvNode* plug_bundle_uri = lilv_plugin_get_bundle_uri(plug);
 	TEST_ASSERT(!strcmp(lilv_node_as_string(plug_bundle_uri), bundle_dir_uri));
 
 	const LilvNodes* data_uris = lilv_plugin_get_data_uris(plug);
 	TEST_ASSERT(lilv_nodes_size(data_uris) == 2);
 
+	LilvNode* project = lilv_plugin_get_project(plug);
+	TEST_ASSERT(!project);
+
 	char* manifest_uri = (char*)malloc(TEST_PATH_MAX);
 	char* data_uri     = (char*)malloc(TEST_PATH_MAX);
 	snprintf(manifest_uri, TEST_PATH_MAX, "%s%s",
@@ -591,6 +616,10 @@ test_plugin(void)
 	TEST_ASSERT(lilv_nodes_contains(data_uris, data_uri_val));
 	lilv_node_free(data_uri_val);
 
+	LilvNode* unknown_uri_val = lilv_new_uri(world, "http://example.org/unknown");
+	TEST_ASSERT(!lilv_nodes_contains(data_uris, unknown_uri_val));
+	lilv_node_free(unknown_uri_val);
+
 	free(manifest_uri);
 	free(data_uri);
 
@@ -631,6 +660,7 @@ test_plugin(void)
 
 	TEST_ASSERT(latency_port);
 	TEST_ASSERT(lilv_port_get_index(plug, latency_port) == 2);
+	TEST_ASSERT(lilv_node_is_blank(lilv_port_get_node(plug, latency_port)));
 
 	LilvNode* rt_feature = lilv_new_uri(world,
 			"http://lv2plug.in/ns/lv2core#hardRTCapable");
@@ -687,7 +717,7 @@ test_plugin(void)
 	char*       blank_tok = lilv_node_get_turtle_token(blank);
 	TEST_ASSERT(!strncmp(blank_tok, "_:", 2));
 	TEST_ASSERT(!strcmp(blank_tok + 2, blank_str));
-	free(blank_tok);
+	lilv_free(blank_tok);
 	lilv_node_free(blank_p);
 	lilv_nodes_free(blanks);
 
@@ -743,6 +773,253 @@ test_plugin(void)
 
 /*****************************************************************************/
 
+static int
+test_project(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
+			BUNDLE_PREFIXES
+			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
+			PLUGIN_NAME("Test plugin with project") " ; "
+			LICENSE_GPL " ; "
+			"lv2:project [ "
+			"  doap:maintainer [ "
+			"    foaf:name \"David Robillard\" ; "
+			"    foaf:homepage <http://drobilla.net> ; foaf:mbox <mailto:d@drobilla.net> ] ; "
+			"] ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 0 ; lv2:symbol \"foo\" ; lv2:name \"bar\" ; "
+			"  lv2:minimum -1.0 ; lv2:maximum 1.0 ; lv2:default 0.5 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 1 ; lv2:symbol \"bar\" ; lv2:name \"Baz\" ; "
+			"  lv2:minimum -2.0 ; lv2:maximum 2.0 ; lv2:default 1.0 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:OutputPort ; "
+			"  lv2:index 2 ; lv2:symbol \"latency\" ; lv2:name \"Latency\" ; "
+			"  lv2:portProperty lv2:reportsLatency ; "
+			"  lv2:designation lv2:latency "
+			"] . \n"
+			":thing doap:name \"Something else\" .\n"))
+		return 0;
+
+	init_uris();
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin* plug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	TEST_ASSERT(plug);
+
+	LilvNode* author_name = lilv_plugin_get_author_name(plug);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(author_name), "David Robillard"));
+	lilv_node_free(author_name);
+
+	LilvNode* author_email = lilv_plugin_get_author_email(plug);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(author_email), "mailto:d@drobilla.net"));
+	lilv_node_free(author_email);
+
+	LilvNode* author_homepage = lilv_plugin_get_author_homepage(plug);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(author_homepage), "http://drobilla.net"));
+	lilv_node_free(author_homepage);
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+static int
+test_no_author(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
+			BUNDLE_PREFIXES
+			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
+			PLUGIN_NAME("Test plugin with project") " ; "
+			LICENSE_GPL " ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 0 ; lv2:symbol \"foo\" ; lv2:name \"bar\" ; "
+			"  lv2:minimum -1.0 ; lv2:maximum 1.0 ; lv2:default 0.5 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 1 ; lv2:symbol \"bar\" ; lv2:name \"Baz\" ; "
+			"  lv2:minimum -2.0 ; lv2:maximum 2.0 ; lv2:default 1.0 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:OutputPort ; "
+			"  lv2:index 2 ; lv2:symbol \"latency\" ; lv2:name \"Latency\" ; "
+			"  lv2:portProperty lv2:reportsLatency ; "
+			"  lv2:designation lv2:latency "
+			"] . \n"
+			":thing doap:name \"Something else\" .\n"))
+		return 0;
+
+	init_uris();
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin* plug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	TEST_ASSERT(plug);
+
+	LilvNode* author_name = lilv_plugin_get_author_name(plug);
+	TEST_ASSERT(!author_name);
+
+	LilvNode* author_email = lilv_plugin_get_author_email(plug);
+	TEST_ASSERT(!author_email);
+
+	LilvNode* author_homepage = lilv_plugin_get_author_homepage(plug);
+	TEST_ASSERT(!author_homepage);
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+static int
+test_project_no_author(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
+			BUNDLE_PREFIXES
+			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
+			PLUGIN_NAME("Test plugin with project") " ; "
+			LICENSE_GPL " ; "
+			"lv2:project [ "
+			"  doap:name \"Fake project\" ;"
+			"] ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 0 ; lv2:symbol \"foo\" ; lv2:name \"bar\" ; "
+			"  lv2:minimum -1.0 ; lv2:maximum 1.0 ; lv2:default 0.5 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 1 ; lv2:symbol \"bar\" ; lv2:name \"Baz\" ; "
+			"  lv2:minimum -2.0 ; lv2:maximum 2.0 ; lv2:default 1.0 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:OutputPort ; "
+			"  lv2:index 2 ; lv2:symbol \"latency\" ; lv2:name \"Latency\" ; "
+			"  lv2:portProperty lv2:reportsLatency ; "
+			"  lv2:designation lv2:latency "
+			"] . \n"
+			":thing doap:name \"Something else\" .\n"))
+		return 0;
+
+	init_uris();
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin* plug = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	TEST_ASSERT(plug);
+
+	LilvNode* author_name = lilv_plugin_get_author_name(plug);
+	TEST_ASSERT(!author_name);
+
+	LilvNode* author_email = lilv_plugin_get_author_email(plug);
+	TEST_ASSERT(!author_email);
+
+	LilvNode* author_homepage = lilv_plugin_get_author_homepage(plug);
+	TEST_ASSERT(!author_homepage);
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+static int
+test_preset(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":plug a lv2:Plugin ; lv2:binary <foo" SHLIB_EXT "> ; rdfs:seeAlso <plugin.ttl> .\n",
+			BUNDLE_PREFIXES
+			":plug a lv2:Plugin ; a lv2:CompressorPlugin ; "
+			PLUGIN_NAME("Test plugin with project") " ; "
+			LICENSE_GPL " ; "
+			"lv2:project [ "
+			"  doap:name \"Fake project\" ;"
+			"] ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 0 ; lv2:symbol \"foo\" ; lv2:name \"bar\" ; "
+			"  lv2:minimum -1.0 ; lv2:maximum 1.0 ; lv2:default 0.5 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 1 ; lv2:symbol \"bar\" ; lv2:name \"Baz\" ; "
+			"  lv2:minimum -2.0 ; lv2:maximum 2.0 ; lv2:default 1.0 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:OutputPort ; "
+			"  lv2:index 2 ; lv2:symbol \"latency\" ; lv2:name \"Latency\" ; "
+			"  lv2:portProperty lv2:reportsLatency ; "
+			"  lv2:designation lv2:latency "
+			"] . \n"
+			"<http://example.org/preset> a pset:Preset ;"
+			"  lv2:appliesTo :plug ;"
+	                  "  rdfs:label \"some preset\" .\n"))
+		return 0;
+
+	init_uris();
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin*  plug    = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	TEST_ASSERT(plug);
+
+	LilvNode*  pset_Preset = lilv_new_uri(world, LV2_PRESETS__Preset);
+	LilvNodes* related     = lilv_plugin_get_related(plug, pset_Preset);
+
+	TEST_ASSERT(lilv_nodes_size(related) == 1);
+
+	lilv_node_free(pset_Preset);
+	lilv_nodes_free(related);
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
+static int
+test_prototype(void)
+{
+	if (!start_bundle(MANIFEST_PREFIXES
+			":prot a lv2:PluginBase ; rdfs:seeAlso <plugin.ttl> .\n"
+			":plug a lv2:Plugin ; lv2:binary <inst" SHLIB_EXT "> ; lv2:prototype :prot .\n",
+			BUNDLE_PREFIXES
+			":prot a lv2:Plugin ; a lv2:CompressorPlugin ; "
+			LICENSE_GPL " ; "
+			"lv2:project [ "
+			"  doap:name \"Fake project\" ;"
+			"] ; "
+			"lv2:port [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 0 ; lv2:symbol \"foo\" ; lv2:name \"bar\" ; "
+			"  lv2:minimum -1.0 ; lv2:maximum 1.0 ; lv2:default 0.5 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:InputPort ; "
+			"  lv2:index 1 ; lv2:symbol \"bar\" ; lv2:name \"Baz\" ; "
+			"  lv2:minimum -2.0 ; lv2:maximum 2.0 ; lv2:default 1.0 "
+			"] , [ "
+			"  a lv2:ControlPort ; a lv2:OutputPort ; "
+			"  lv2:index 2 ; lv2:symbol \"latency\" ; lv2:name \"Latency\" ; "
+			"  lv2:portProperty lv2:reportsLatency ; "
+			"  lv2:designation lv2:latency "
+			"] . \n"
+			":plug doap:name \"Instance\" .\n"))
+		return 0;
+
+	init_uris();
+	const LilvPlugins* plugins = lilv_world_get_all_plugins(world);
+	const LilvPlugin*  plug    = lilv_plugins_get_by_uri(plugins, plugin_uri_value);
+	TEST_ASSERT(plug);
+
+	// Test non-inherited property
+	LilvNode* name = lilv_plugin_get_name(plug);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(name), "Instance"));
+	lilv_node_free(name);
+
+	// Test inherited property
+	const LilvNode* binary = lilv_plugin_get_library_uri(plug);
+	TEST_ASSERT(strstr(lilv_node_as_string(binary), "inst" SHLIB_EXT));
+
+	cleanup_uris();
+	return 1;
+}
+
+/*****************************************************************************/
+
 static int
 test_port(void)
 {
@@ -799,6 +1076,12 @@ test_port(void)
 	TEST_ASSERT(p3 == NULL);
 	lilv_node_free(nopsym);
 
+	// Try getting an invalid property
+	LilvNode*  num     = lilv_new_int(world, 1);
+	LilvNodes* nothing = lilv_port_get_value(plug, p, num);
+	TEST_ASSERT(!nothing);
+	lilv_node_free(num);
+
 	LilvNode* audio_class = lilv_new_uri(world,
 			"http://lv2plug.in/ns/lv2core#AudioPort");
 	LilvNode* control_class = lilv_new_uri(world,
@@ -855,6 +1138,12 @@ test_port(void)
 	TEST_ASSERT(!strcmp(lilv_node_as_string(name), "store"));
 	lilv_node_free(name);
 
+	// Invalid language
+	setenv("LANG", "1!", 1);
+	name = lilv_port_get_name(plug, p);
+	TEST_ASSERT(!strcmp(lilv_node_as_string(name), "store"));
+	lilv_node_free(name);
+
 	setenv("LANG", "en_CA.utf-8", 1);
 
 	// Language tagged value with no untranslated values
@@ -1000,6 +1289,7 @@ test_port(void)
 	lilv_scale_points_free(points);
 	lilv_node_free(control_class);
 	lilv_node_free(audio_class);
+	lilv_node_free(out_class);
 	lilv_node_free(in_class);
 	cleanup_uris();
 	return 1;
@@ -1085,11 +1375,15 @@ test_ui(void)
 	LilvNode* ui_class_uri = lilv_new_uri(world,
 			"http://lv2plug.in/ns/extensions/ui#GtkUI");
 
+	LilvNode* unknown_ui_class_uri = lilv_new_uri(world,
+			"http://example.org/mysteryUI");
+
 	TEST_ASSERT(lilv_node_equals(lilv_nodes_get_first(classes), ui_class_uri));
 	TEST_ASSERT(lilv_ui_is_a(ui0, ui_class_uri));
 
 	const LilvNode* ui_type = NULL;
 	TEST_ASSERT(lilv_ui_is_supported(ui0, ui_supported, ui_class_uri, &ui_type));
+	TEST_ASSERT(!lilv_ui_is_supported(ui0, ui_supported, unknown_ui_class_uri, &ui_type));
 	TEST_ASSERT(lilv_node_equals(ui_type, ui_class_uri));
 
 	const LilvNode* plug_bundle_uri = lilv_plugin_get_bundle_uri(plug);
@@ -1106,6 +1400,7 @@ test_ui(void)
 	TEST_ASSERT(lilv_node_equals(expected_uri, ui_binary_uri));
 
 	free(ui_binary_uri_str);
+	lilv_node_free(unknown_ui_class_uri);
 	lilv_node_free(ui_class_uri);
 	lilv_node_free(ui_uri);
 	lilv_node_free(ui2_uri);
@@ -1262,6 +1557,12 @@ test_state(void)
 	// Ensure they are equal
 	TEST_ASSERT(lilv_state_equals(state, state2));
 
+	// Check that we can't delete unsaved state
+	TEST_ASSERT(lilv_state_delete(world, state));
+
+	// Check that state has no URI
+	TEST_ASSERT(!lilv_state_get_uri(state));
+
 	// Check that we can't save a state with no URI
 	char* bad_state_str = lilv_state_to_string(
 		world, &map, &unmap, state, NULL, NULL);
@@ -1270,7 +1571,7 @@ test_state(void)
 	// Check that we can't restore the NULL string (and it doesn't crash)
 	LilvState* bad_state = lilv_state_new_from_string(world, &map, NULL);
 	TEST_ASSERT(!bad_state);
-	
+
 	// Save state to a string
 	char* state1_str = lilv_state_to_string(
 		world, &map, &unmap, state, "http://example.org/state1", NULL);
@@ -1341,6 +1642,19 @@ test_state(void)
 
 	LilvState* state6 = lilv_state_new_from_world(world, &map, test_state_node);
 	TEST_ASSERT(lilv_state_equals(state, state6));  // Round trip accuracy
+
+	// Check that loaded state has correct URI
+	TEST_ASSERT(lilv_state_get_uri(state6));
+	TEST_ASSERT(!strcmp(lilv_node_as_string(lilv_state_get_uri(state6)),
+	                    state_uri));
+
+	lilv_world_unload_resource(world, test_state_node);
+	lilv_world_unload_bundle(world, test_state_bundle);
+
+	LilvState* state6_2 = lilv_state_new_from_world(world, &map, test_state_node);
+	TEST_ASSERT(!state6_2);  // No longer present
+	lilv_state_free(state6_2);
+
 	lilv_node_free(test_state_bundle);
 	lilv_node_free(test_state_node);
 
@@ -1366,6 +1680,15 @@ test_state(void)
 	lilv_instance_connect_port(instance, 1, &out);
 	lilv_instance_run(instance, 1);
 
+	// Test instantiating twice
+	LilvInstance* instance2 = lilv_plugin_instantiate(plugin, 48000.0, ffeatures);
+	if (!instance2) {
+		fatal_error("Failed to create multiple instances of <%s>\n",
+		            lilv_node_as_uri(state_plugin_uri));
+		return 0;
+	}
+	lilv_instance_free(instance2);
+
 	// Get instance state state
 	LilvState* fstate = lilv_state_new_from_instance(
 		plugin, instance, &map,
@@ -1443,6 +1766,9 @@ test_state(void)
 	TEST_ASSERT(lilv_state_equals(fstate72, fstate7));
 	TEST_ASSERT(!lilv_state_equals(fstate6, fstate72));
 
+	// Delete saved state
+	lilv_state_delete(world, fstate7);
+
 	lilv_instance_deactivate(instance);
 	lilv_instance_free(instance);
 
@@ -1585,9 +1911,38 @@ test_string(void)
 
 /*****************************************************************************/
 
+static int
+test_world(void)
+{
+	if (!init_world()) {
+		return 0;
+	}
+
+	LilvNode* num = lilv_new_int(world, 4);
+	LilvNode* uri = lilv_new_uri(world, "http://example.org/object");
+
+	LilvNodes* matches = lilv_world_find_nodes(world, num, NULL, NULL);
+	TEST_ASSERT(!matches);
+
+	matches = lilv_world_find_nodes(world, NULL, num, NULL);
+	TEST_ASSERT(!matches);
+
+	matches = lilv_world_find_nodes(world, NULL, uri, NULL);
+	TEST_ASSERT(!matches);
+
+	lilv_node_free(uri);
+	lilv_node_free(num);
+
+	lilv_world_unload_bundle(world, NULL);
+
+	return 1;
+}
+
+/*****************************************************************************/
+
 /* add tests here */
 static struct TestCase tests[] = {
-	TEST_CASE(utils),
+	TEST_CASE(util),
 	TEST_CASE(value),
 	TEST_CASE(verify),
 	TEST_CASE(no_verify),
@@ -1595,12 +1950,18 @@ static struct TestCase tests[] = {
 	TEST_CASE(lv2_path),
 	TEST_CASE(classes),
 	TEST_CASE(plugin),
+	TEST_CASE(project),
+	TEST_CASE(no_author),
+	TEST_CASE(project_no_author),
+	TEST_CASE(preset),
+	TEST_CASE(prototype),
 	TEST_CASE(port),
 	TEST_CASE(ui),
 	TEST_CASE(bad_port_symbol),
 	TEST_CASE(bad_port_index),
 	TEST_CASE(bad_port_index),
 	TEST_CASE(string),
+	TEST_CASE(world),
 	TEST_CASE(state),
 	{ NULL, NULL }
 };
